﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Problems.Robocode {
  [StorableClass]
  [Creatable("Problems")]
  [Item("Robocode Problem",
        "The Robocode problem for genetic programming.")]
  public class RobocodeProblem : SingleObjectiveHeuristicOptimizationProblem<RobocodeEvaluator,
    ISymbolicExpressionTreeCreator> {
    #region parameter names
    private const string TankProgramParameterName = "TankProgram";
    private const string MaxTankProgramLengthParameterName = "MaxProgramLength";
    private const string MaxTankProgramDepthParameterName = "MaxProgramDepth";
    private const string TankGrammarParameterName = "Grammar";
    private const string RobocodePathParamaterName = "RobocodePath";
    private const string NrOfRoundsParameterName = "NrOfRounds";
    private const string EnemiesParameterName = "Enemies";
    #endregion

    #region Parameters
    public IFixedValueParameter<IntValue> MaxTankProgramLengthParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[MaxTankProgramLengthParameterName]; }
    }
    public IFixedValueParameter<IntValue> MaxTankProgramDepthParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[MaxTankProgramDepthParameterName]; }
    }
    public IValueParameter<Grammar> GrammarParameter {
      get { return (IValueParameter<Grammar>)Parameters[TankGrammarParameterName]; }
    }
    public IFixedValueParameter<DirectoryValue> RobocodePathParameter {
      get { return (IFixedValueParameter<DirectoryValue>)Parameters[RobocodePathParamaterName]; }
    }
    public IFixedValueParameter<IntValue> NrOfRoundsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[NrOfRoundsParameterName]; }
    }
    public IValueParameter<ICheckedItemList<StringValue>> EnemiesParameter {
      get { return (IValueParameter<ICheckedItemList<StringValue>>)Parameters[EnemiesParameterName]; }
    }
    #endregion

    [StorableConstructor]
    protected RobocodeProblem(bool deserializing)
      : base(deserializing) {
    }
    protected RobocodeProblem(RobocodeProblem original, Cloner cloner)
      : base(original, cloner) {
    }

    public RobocodeProblem()
      : base(new RobocodeEvaluator(), new RampedHalfAndHalfTreeCreator()) {
      DirectoryValue robocodeDir = new DirectoryValue();
      robocodeDir.Value = @"C:\robocode";

      var robotsDir = new DirectoryInfo(Path.Combine(robocodeDir.Value, "robots"));
      var robotFiles = robotsDir.GetFiles("*", SearchOption.AllDirectories)
                                .Where(x => x.Extension == ".class" || x.Extension == ".jar");
      var robotSet = new Dictionary<string, StringValue>();
      foreach (var robot in robotFiles) {
        string robotName = Path.Combine(robot.DirectoryName, Path.GetFileNameWithoutExtension(robot.FullName));
        robotName = robotName.Remove(0, robotsDir.FullName.Length);
        string[] nameParts = robotName.Split(new[] { Path.DirectorySeparatorChar }, StringSplitOptions.RemoveEmptyEntries);
        robotName = string.Join(".", nameParts);
        robotSet[robot.FullName] = new StringValue(robotName);
      }
      var robotList = new CheckedItemList<StringValue>(robotSet.Values);

      Parameters.Add(new FixedValueParameter<IntValue>(MaxTankProgramDepthParameterName, "Maximal depth of the Robocode tank program.", new IntValue(6)));
      Parameters.Add(new FixedValueParameter<IntValue>(MaxTankProgramLengthParameterName, "Maximal length of the tank program.", new IntValue(1000)));
      Parameters.Add(new ValueParameter<Grammar>(TankGrammarParameterName, "Grammar for the tank program.", new Grammar()));
      Parameters.Add(new FixedValueParameter<DirectoryValue>(RobocodePathParamaterName, "Path of the Robocode installation.", robocodeDir));
      Parameters.Add(new FixedValueParameter<IntValue>(NrOfRoundsParameterName, "Nr. of Rounds a Robot has to fight against each opponent.", new IntValue(3)));
      Parameters.Add(new ValueParameter<ICheckedItemList<StringValue>>(EnemiesParameterName, "The enemies that should be battled.", robotList));

      Maximization.Value = true;
      InitializeOperators();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RobocodeProblem(this, cloner);
    }

    private void InitializeOperators() {
      Operators.AddRange(
        ApplicationManager.Manager.GetInstances<ISymbolicExpressionTreeOperator>());
      Operators.Add(new MinAverageMaxSymbolicExpressionTreeLengthAnalyzer());
      Operators.Add(new SymbolicExpressionSymbolFrequencyAnalyzer());
      Operators.Add(new BestSolutionAnalyzer());
      ParameterizeOperators();
      ParameterizeAnalyzers();
    }

    protected override void OnEvaluatorChanged() {
      base.OnEvaluatorChanged();
      Evaluator.TankProgramParameter.ActualName =
        TankProgramParameterName;
      ParameterizeAnalyzers();
      ParameterizeOperators();
    }

    protected override void OnSolutionCreatorChanged() {
      base.OnSolutionCreatorChanged();
      SolutionCreator.SymbolicExpressionTreeParameter.ActualName =
        TankProgramParameterName;
      ParameterizeAnalyzers();
      ParameterizeOperators();
    }

    private void ParameterizeAnalyzers() {
      var analyzers = Operators.OfType<IAnalyzer>();
      foreach (var o in analyzers.OfType<ISymbolicExpressionTreeAnalyzer>()) {
        o.SymbolicExpressionTreeParameter.ActualName =
          SolutionCreator.SymbolicExpressionTreeParameter.ActualName;
      }
      foreach (var o in analyzers.OfType<BestSolutionAnalyzer>()) {
        o.QualityParameter.ActualName = Evaluator.QualityParameter.ActualName;
      }
    }

    private void ParameterizeOperators() {
      var operators = Parameters
        .OfType<IValueParameter>()
        .Select(p => p.Value)
        .OfType<IOperator>()
        .Union(Operators);
      foreach (var o in operators.OfType<ISymbolicExpressionTreeGrammarBasedOperator>()) {
        o.SymbolicExpressionTreeGrammarParameter.ActualName =
          TankGrammarParameterName;
      }
      foreach (var o in operators.OfType<ISymbolicExpressionTreeSizeConstraintOperator>()) {
        o.MaximumSymbolicExpressionTreeDepthParameter.ActualName =
          MaxTankProgramDepthParameterName;
        o.MaximumSymbolicExpressionTreeLengthParameter.ActualName =
          MaxTankProgramLengthParameterName;
      }
      foreach (var op in operators.OfType<RobocodeEvaluator>()) {
        op.TankProgramParameter.ActualName =
          SolutionCreator.SymbolicExpressionTreeParameter.ActualName;
      }
      foreach (var op in operators.OfType<ISymbolicExpressionTreeCrossover>()) {
        op.ParentsParameter.ActualName =
          SolutionCreator.SymbolicExpressionTreeParameter.ActualName;
        op.ChildParameter.ActualName =
          SolutionCreator.SymbolicExpressionTreeParameter.ActualName;
      }
      foreach (var op in operators.OfType<ISymbolicExpressionTreeManipulator>()) {
        op.SymbolicExpressionTreeParameter.ActualName =
          SolutionCreator.SymbolicExpressionTreeParameter.ActualName;
      }
      foreach (var op in operators.OfType<ISymbolicExpressionTreeCreator>()) {
        op.SymbolicExpressionTreeParameter.ActualName =
          SolutionCreator.SymbolicExpressionTreeParameter.ActualName;
      }
    }
  }
}